@@ -3,8 +3,10 @@
 
 ((window) => {
   const core = window.Deno.core;
-  const { BadResource, Interrupted } = core;
+  const { BadResourcePrototype, InterruptedPrototype } = core;
+  const { ReadableStream, WritableStream } = window.__bootstrap.streams;
   const {
+    ObjectPrototypeIsPrototypeOf,
     PromiseResolve,
     SymbolAsyncIterator,
     Uint8Array,
@@ -58,10 +60,75 @@
     return core.opAsync("op_dns_resolve", { query, recordType, options });
   }
 
+  const DEFAULT_CHUNK_SIZE = 16_640;
+
+  function tryClose(rid) {
+    try {
+      core.close(rid);
+    } catch {
+      // Ignore errors
+    }
+  }
+
+  function readableStreamForRid(rid) {
+    return new ReadableStream({
+      type: "bytes",
+      async pull(controller) {
+        const v = controller.byobRequest.view;
+        try {
+          const bytesRead = await read(rid, v);
+          if (bytesRead === null) {
+            tryClose(rid);
+            controller.close();
+            controller.byobRequest.respond(0);
+          } else {
+            controller.byobRequest.respond(bytesRead);
+          }
+        } catch (e) {
+          controller.error(e);
+          tryClose(rid);
+        }
+      },
+      cancel() {
+        tryClose(rid);
+      },
+      autoAllocateChunkSize: DEFAULT_CHUNK_SIZE,
+    });
+  }
+
+  function writableStreamForRid(rid) {
+    return new WritableStream({
+      async write(chunk, controller) {
+        try {
+          let nwritten = 0;
+          while (nwritten < chunk.length) {
+            nwritten += await write(
+              rid,
+              TypedArrayPrototypeSubarray(chunk, nwritten),
+            );
+          }
+        } catch (e) {
+          controller.error(e);
+          tryClose(rid);
+        }
+      },
+      close() {
+        tryClose(rid);
+      },
+      abort() {
+        tryClose(rid);
+      },
+    });
+  }
+
   class Conn {
     #rid = 0;
     #remoteAddr = null;
     #localAddr = null;
+
+    #readable;
+    #writable;
+
     constructor(rid, remoteAddr, localAddr) {
       this.#rid = rid;
       this.#remoteAddr = remoteAddr;
@@ -95,6 +162,28 @@
     closeWrite() {
       return shutdown(this.rid);
     }
+
+    setNoDelay(nodelay = true) {
+      return core.opSync("op_set_nodelay", this.rid, nodelay);
+    }
+
+    setKeepAlive(keepalive = true) {
+      return core.opSync("op_set_keepalive", this.rid, keepalive);
+    }
+
+    get readable() {
+      if (this.#readable === undefined) {
+        this.#readable = readableStreamForRid(this.rid);
+      }
+      return this.#readable;
+    }
+
+    get writable() {
+      if (this.#writable === undefined) {
+        this.#writable = writableStreamForRid(this.rid);
+      }
+      return this.#writable;
+    }
   }
 
   class Listener {
@@ -124,7 +213,10 @@
       try {
         conn = await this.accept();
       } catch (error) {
-        if (error instanceof BadResource || error instanceof Interrupted) {
+        if (
+          ObjectPrototypeIsPrototypeOf(BadResourcePrototype, error) ||
+          ObjectPrototypeIsPrototypeOf(InterruptedPrototype, error)
+        ) {
           return { value: undefined, done: true };
         }
         throw error;
@@ -191,7 +283,10 @@
         try {
           yield await this.receive();
         } catch (err) {
-          if (err instanceof BadResource || err instanceof Interrupted) {
+          if (
+            ObjectPrototypeIsPrototypeOf(BadResourcePrototype, err) ||
+            ObjectPrototypeIsPrototypeOf(InterruptedPrototype, err)
+          ) {
             break;
           }
           throw err;
@@ -237,4 +332,8 @@
     Datagram,
     resolveDns,
   };
+  window.__bootstrap.streamUtils = {
+    readableStreamForRid,
+    writableStreamForRid,
+  };
 })(this);
